#include "ep_lib.hpp"
#include <mpi.h>
#include "ep_declaration.hpp"
#include "ep_mpi.hpp"

using namespace std;


namespace ep_lib 
{

  int MPI_Intercomm_merge(MPI_Comm inter_comm, bool high, MPI_Comm *newintracomm)
  {
    
    int ep_rank = inter_comm->ep_comm_ptr->size_rank_info[0].first;
    int ep_size = inter_comm->ep_comm_ptr->size_rank_info[0].second;
    int ep_rank_loc = inter_comm->ep_comm_ptr->size_rank_info[1].first;
    int num_ep = inter_comm->ep_comm_ptr->size_rank_info[1].second;
    int mpi_rank = inter_comm->ep_comm_ptr->size_rank_info[2].first;

    int remote_ep_size = inter_comm->inter_rank_map->size();

    int new_ep_rank = high? remote_ep_size + ep_rank : ep_rank;


    MPI_Comm_dup(inter_comm, newintracomm);

    (*newintracomm)->is_intercomm = false;
    (*newintracomm)->inter_rank_map->clear();
    delete (*newintracomm)->inter_rank_map;

    (*newintracomm)->ep_comm_ptr->size_rank_info[0].second = ep_size + remote_ep_size;
    
    ///////////////////////////////////
    int int_high = high? 1 : 0;
    int sum_high;
    
    MPI_Allreduce(&int_high, &sum_high, 1, MPI_INT, MPI_SUM, *newintracomm);

    if(sum_high==0 || sum_high==ep_size+remote_ep_size)
    {
      printf("MPI_Intercomm_merge error: please define high with different value...\n");
      MPI_Abort(inter_comm, 0);
    }
    
    
    
    ///////////////////////////////////
    
    (*newintracomm)->ep_comm_ptr->size_rank_info[0].first = new_ep_rank;

    // modif ep_rank_map
    int my_triple[3] = {new_ep_rank, ep_rank_loc, mpi_rank};
    int *my_triple_list = new int[3*(ep_size+remote_ep_size)];

    
    MPI_Allgather(my_triple, 3, MPI_INT, my_triple_list, 3, MPI_INT, *newintracomm); 


#ifdef _showinfo
    for(int i=0; i<ep_size+remote_ep_size; i++)
    {
      if(new_ep_rank == i)
      {
        for(int j=0; j<ep_size+remote_ep_size; j++)
        {
          printf("rank %d : my_triple_list[%d] = %d %d %d\n", i, j, my_triple_list[3*j], my_triple_list[3*j+1], my_triple_list[3*j+2]);
        }
        printf("\n");
      }
      MPI_Barrier(*newintracomm);
    }
#endif

    if((*newintracomm)->ep_comm_ptr->size_rank_info[1].first==0)
    {
      for(int i=0; i<ep_size+remote_ep_size; i++)
      {
        (*newintracomm)->ep_comm_ptr->comm_list[0]->ep_rank_map->at(my_triple_list[3*i]) = std::make_pair(my_triple_list[3*i+1], my_triple_list[3*i+2]);
      }
    }

    MPI_Barrier_local(*newintracomm);

#ifdef _showinfo

    for(int i=0; i<ep_size+remote_ep_size; i++)
    {
      if(new_ep_rank == i)
      {
        for(EP_RANK_MAP::iterator it = (*newintracomm)->ep_rank_map->begin(); it != (*newintracomm)->ep_rank_map->end(); it++)
        {
          printf("rank %d : ep_rank_map[%d] = %d %d\n", i, it->first, it->second.first, it->second.second);
        }
        printf("\n");
      }
      MPI_Barrier(*newintracomm);
    }
#endif

    delete[] my_triple_list;

  }
  
}
